from kp_fraydit.producers.producer import Producer
from kp_fraydit.schema_client import SchemaEngine

class LogProducer(Producer):
    def __init__(self):

        # Create the topic if not found
        seng = SchemaEngine()
        try:
            seng.get_latest_schema('logger')
        except:
            attempts = 0
            while True:
                if seng.create_topic: break
                if attempts == 10:
                    raise
            
            # Create the schema
            key_schema = ''
            value_schema = ''
            seng.register_schema('logger', key_schema)

        

        super().__init__('logger')


def read_from_pickle():
    import pickle
    objects = []

    with open('save_test1.pkl', 'rb') as f:
        try:
            while True:
                objects.append(pickle.load(f))
        except EOFError:
            pass


from confluent_kafka.admin import AdminClient
from confluent_kafka.admin import NewTopic

from kp_fraydit.connections.connection import KafkaConnection
from kp_fraydit.metaclasses import SingletonMeta


kConn = KafkaConnection()


class AdminEngine(AdminClient, metaclass=SingletonMeta):
    def __init__(self):
        # my_server = kConn.kafka_registry_listener.split("/")[2]
        ip, port = kConn.kafka_broker_listener.split(':')
        self.__conf = { "bootstrap.servers": f'{ip}:{port}' }
        super().__init__(self.__conf)

    def topic_exists(self, topic_name):
        topic_metadata = self.list_topics()
        if topic_metadata.topics.get(topic_name) is None: return False
        return True

    def create_topic(self, topic_name, num_partitions=5, retention_time=-1, retention_size=-1):
        if not self.topic_exists(topic_name):
            my_topic = NewTopic(topic=topic_name, num_partitions=5, config={'retention.ms': f'{retention_time}', 'retention.bytes': f'{retention_size}', })
            self.create_topics([my_topic,])



import socket
import time
import abc
import queue
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
from multiprocessing import Process
from threading import Thread
from queue import Queue
import os

from confluent_kafka import SerializingProducer, KafkaException

from kp_fraydit import custom_errors
from kp_fraydit.custom_types import check_int, key_exists
from kp_fraydit.metaclasses import SingletonMeta


# END IMPORTS ///////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# HELPER FUNCTIONS //////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


def check_host_port(host: str, port: int, tag='', tries=3, sleep_interval=1):
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    a_socket.settimeout(3)
    location = (host, port)
    actual_tries = 0
    while tries > actual_tries:
        
        try:
            result_of_check = a_socket.connect_ex(location)
            if result_of_check == 0: return True, host, port, tag, time.time()
            actual_tries += 1
            time.sleep(sleep_interval)
        
        except:
            actual_tries += 1
            time.sleep(sleep_interval)
                
    return False, host, port, tag


def get_ip_and_port_from_string(ip_string: str):
    
    stripped_string = ip_string.strip()
    if stripped_string[:7] == 'http://': stripped_string = ip_string[7:]
    if stripped_string[:8] == 'https://': stripped_string = ip_string[8:]
    
    if len(stripped_string.split(':')) > 1: return stripped_string.split(':')[0], stripped_string.split(':')[1]
    else: return stripped_string, ''


def valid_address_format(url):
    stripped_string = url.strip()
    port = stripped_string.split(':')[-1:][0]
    if not check_int(port): return False
    
    if stripped_string[:7] == 'http://' or stripped_string[:8] == 'https://': 
        return True
    
    return False


# END HELPER FUNCTIONS //////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# PRODUCERNATIVE AND HELPER CLASSES /////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


'''
The actual class that is used to do the produccing. It is transactional and is built to ensure an Exact Once Semantics. 
It is used by the KafkaConnection only and in a threaded manner.
'''

class ProducerNative(SerializingProducer):
    def __init__(self, conf_, value_schema_name, key_schema_name, tx_id):
        self.tx_id = tx_id
        conf_['transactional.id'] = tx_id
        self.__conf = conf_
        self.__value_schema_name = value_schema_name
        self.__key_schema_name = key_schema_name
        super().__init__(conf_)
        # self.queue = Queue()    


    def produce(self, topic, values, keys): # Overrides parent
        self.init_transactions()

        '''
        The infinite loop tries a transaction as long as the error is not abortable. If it is abortable, it is then requeued to the connection
        '''
        while True:
            try:
                self.begin_transaction()
                super().produce(topic, keys, values)
                self.commit_transaction()
                self.flush()
                break
            
            except KafkaException as e:
                if e.args[0].retrievable():
                    continue # retry the transaction until an abortable failure happens
                    '''
                    Upon transaction failure or unidentified failure, the transaction is requeued to be processed
                    '''
                elif e.args[0].txn_requires_abort():
                    self.abort_transaction()
                    KafkaConnection.produce_queue.put([topic, self.__value_schema_name, self.__key_schema_name, values, keys, self.__conf])
                else:
                    self.abort_transaction()
                    KafkaConnection.produce_queue.put([self.__topic_name, self.__value_schema_name, self.__key_schema_name, values, keys, self.__conf])

    
# END PRODUCERNATIVE AND HELPER CLASSES /////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# CONNECTION CLASS //////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


'''
The metaclass that makes the Connection to Kafka a Singleton
'''
# class ConnectionMeta(type):
#     def __init__(cls, name, bases, attrs, **kwargs):
#         super().__init__(name, bases, attrs)
#         cls._instance = None
        

#     def __call__(cls, *args, **kwargs):
#         if cls._instance is None:
#             cls._instance = super().__call__(*args, **kwargs)
        
#         return cls._instance


class KafkaConnection(metaclass=SingletonMeta):

    # INITIALIZATION ////////////////////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    def __init__(self, thread_count=250):
        
        self._kafka_broker_listener = os.environ.get('KAFKA_BROKER_LISTENER')
        if self._kafka_broker_listener is None: self._kafka_broker_listener = "10.100.100.1:9092"
        self._kafka_registry_listener = os.environ.get('KAFKA_REGISTRY_LISTENER')
        if self._kafka_registry_listener is None: self._kafka_registry_listener = "http://10.100.100.1:8081"
        self._status = {}
        self._thread_count = thread_count

        self.__observers = set()
        self.__subject_state = {}
        self. __subject_state['broker_online'] = False
        self.__subject_state['registry_online'] = False
    
        self.loop_active = True
        self.__check_connections()
        self.__produce_queue = queue.Queue()
        self.__start_producer_queue()

    # END INITIALIZATION ////////////////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    # KAFKA CONNECTION ATTRIBUTES AND METHODS ///////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    @property
    def kafka_broker_listener(self):
        return self._kafka_broker_listener


    @kafka_broker_listener.setter
    def kafka_broker_listener(self, value):
        ip, port = get_ip_and_port_from_string(value)
        formatted_address = f'{ip}:{port}'
        os.environ['KAFKA_BROKER_LISTENER'] = formatted_address
        self._kafka_broker_listener = formatted_address
        
        
    @property
    def kafka_registry_listener(self):
        return self._kafka_registry_listener


    @kafka_registry_listener.setter
    def kafka_registry_listener(self, value):
        if not valid_address_format(value):
            print (f'Address supplied ({value}) for kafka registry listener is not valid. Please format with http or https preceding address and a port number')
            return
        os.environ['KAFKA_REGISTRY_LISTENER'] = value
        self._kafka_registry_listener = value

    @property
    def subject_state(self):
        return self.__subject_state


    @subject_state.setter
    def subject_state(self, value):
        
        if not isinstance(value, dict): 
            print ('The subject state must be a dict: "registry_online": bool, "broker_online": bool')
            return
            if value.get('broker_online') is None or value.get('registry_online') is None:
                print ('The subject state must be a dict: "registry_online": bool, "broker_online": bool')
                return
        
        previous_state = self.__subject_state
        self.__subject_state['broker_online'] = value['broker_online']
        self.__subject_state['registry_online'] = value['registry_online']
        if previous_state != self.__subject_state:
            self.__notify()
        

    @property
    def broker_online(self):
        return self.subject_state['broker_online']


    @broker_online.setter
    def broker_online(self, value):
        if not isinstance(value, bool):
            print ('Broker online must be a boolean')
            return
        self.subject_state = {'broker_online': value, 'registry_online': self.__subject_state['registry_online']}


    @property
    def registry_online(self):
        return self.subject_state['registry_online']


    @registry_online.setter
    def registry_online(self, value):
        if not isinstance(value, bool):
            print ('Registry online must be a boolean')
            return
        self.subject_state = {'registry_online': value, 'broker_online': self.__subject_state['broker_online']}


    def __check_conn_loop(self):
        
        '''
        Checks the status of the connection. This runs only when the connection loop_active is true. 
        It is used in check_connections where it is infinitely looped in a daemon thread.
        '''
        while self.loop_active:
            
            changed = False

            a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            a_socket2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            a_socket.settimeout(10)
            a_socket2.settimeout(10)

            # Kafka broker
            ip, port = get_ip_and_port_from_string(self.kafka_broker_listener)
            location = (str(ip), int(port))
            result_of_check = a_socket.connect_ex(location)
            
            if result_of_check == 0: 
                broker_online = True
            else: 
                broker_online = False
            
            # Kafka registry
            ip2, port2 = get_ip_and_port_from_string(self.kafka_registry_listener)
            location2 = (str(ip2), int(port2))
            result_of_check = a_socket2.connect_ex(location2)
            if result_of_check == 0: 
                registry_online = True
            else: 
                registry_online = False
            

            if broker_online != self.broker_online: changed = True
            if registry_online != self.registry_online: changed = True
                        
            self.subject_state = {'broker_online': broker_online, 'registry_online': registry_online}
            
            if changed: 
                self.__notify()

    def __check_connections(self): # Infinite loop that runs as a deamon thread to update the broker and registry status
        t = Thread(target=self.__check_conn_loop, args=(), daemon=True)
        t.start()
    
    @property
    def main_conf(self):
        if kConn.kafka_broker_listener is None: return ""

        ip, port = get_ip_and_port_from_string(kConn.kafka_broker_listener)
        formatted_address = f'{ip}:{port}'
        {'bootstrap.servers': formatted_address,}
    # END KAFKA CONNECTION ATTRIBUTES AND METHODS ///////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    # OBSERVER ATTRIBUTES AND METHODS ///////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    def attach(self, observer): # Attcahes subscribers
        observer._subject = self
        self.__observers.add(observer)
        self.__notify()
        
    def detach(self, observer): # Detaches subscribers
        observer._subject = None
        self.__observers.discard(observer)
        
    def __notify(self): # Notifies subscribers of changes
        
        for observer in self.__observers:
            observer.update(self.subject_state)


    # PRODUCE RECORDS METHODS AND ATTRIBUTES ////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @property
    def produce_queue(self):
        return self.__produce_queue

    @produce_queue.setter
    def produce_queue(self, value):
        self.__produce_queue = value

    @property
    def thread_count(self):
        return self._thread_count

    @thread_count.setter
    def thread_count(self, value):
        self._thread_count = value


    def __produce_record(self, thread_no, topic, value_schema_name, key_schema_name, values, keys, conf_):
        # print(f'Thread #{thread_no} is doing task #{task} in the queue.')
        #print(f'Trying task {topic} | {value_args} | {key_args} | {value_schema} | {key_schema}')

        prod = ProducerNative(conf_, value_schema_name, key_schema_name, thread_no)
        prod.produce(topic, values, keys)
        
        #print(f'Thread #{prod.tx_id} finished task #{thread_no} | {topic} | {value_args} | {key_args} | {value_schema} | {key_schema}')
        print(f'Thread #{prod.tx_id} finished task #{thread_no} | {topic} | {value_schema_name} | {key_schema_name}')


    def __start_producer_queue(self):
        '''
        queue_consumer is an infinite loop that runs only when the broker is connected.
            It then starts a specified number of thread (self.thread_count) that consumes the records that are placed 
            in the queue by instances of the Producers. These threads are passed the records to consume. 
            Once all threads are finished, it loops again.
        '''
        def __queue_consumer():        
            
            while True:        
            
                prods = []
                workers = []
                while self.broker_online:
                    
                    for i in range(self.thread_count):
                        task = self.produce_queue.get()
                        
                        topic, value_schema_name, key_schema_name, values, keys,  conf_ = task
                        prods.append(ProducerNative(conf_=conf_, value_schema_name=value_schema_name, key_schema_name=key_schema_name, tx_id=i))             
                        
                        worker = Thread(target=self.__produce_record, args=(i, topic, value_schema_name, key_schema_name, values, keys, conf_), daemon=False)
                        workers.append(worker)
                        worker.start()

                    for i in range(self.thread_count): 
                        workers[i].join()
                        
        queue_thread = Thread(target=__queue_consumer, args=(), daemon=True)
        queue_thread.start()
            
    # END PRODUCER RECORD METHODS AND ATTRIBUTES ////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# END CONNECTION CLASS //////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# Create the Singleton instance
kConn = KafkaConnection()


import json

from kp_fraydit.connectors.connector import Connector
from kp_fraydit.schema_client import SchemaEngine
from kp_fraydit.admin import AdminEngine


# # conf = {'bootstrap.servers': formatted_address,}
eng = SchemaEngine()
admin = AdminEngine()


class PriceConnector(Connector):
    def __init__(self, source_name):
        super().__init__(source_name)
        
        self.__source_name = source_name
        self.__value_schema_name = f'{source_name}-value'
        self.__key_schema_name = f'{source_name}-key'
        
        # Validate the schemas
        eng.validate_schema(self.__source_name, self.__value_schema_name, 'kp_fraydit/financial/price_data_value.avro')
        eng.validate_schema(self.__source_name, self.__key_schema_name, 'kp_fraydit/financial/price_data_key.avro')


    @property
    def source_name(self):
        return self.__source_name

    @property
    def value_schema_name(self):
        return self.__value_schema_name

    @property
    def key_schema_name(self):
        return self.__key_schema_name

    def create_schema(self, schema_dict):
        pass


from datetime import datetime, timezone
import pytz

def now_as_string():        
    now = datetime.now(timezone.utc)
    return now.strftime("%Y-%m-%d %H:%M:%S")

def today_as_string():        
    now = datetime.now(timezone.utc)
    return now.strftime("%Y-%m-%d")

def utc_now_as_long():
    now = datetime.now(tz=timezone.utc)
    new_val = now.strftime("%Y%m%d%H%M%S")
    return int(new_val)

def newyork_now_as_long():
    now = datetime.now(tz=pytz.timezone('America/New_York'))
    new_val = now.strftime("%Y%m%d%H%M%S")
    return int(new_val)





class SingletonMeta(type):
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None
        

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        
        return cls._instance




import abc

class Observer(metaclass=abc.ABCMeta):
    """
    Define an updating interface for objects that should be notified of
    changes in a subject.
    """

    def __init__(self):
        self._subject = None
        self._observer_state = None

    @abc.abstractmethod
    def update(self, arg):
        pass


class ConcreteObserver(Observer):
    """
    Implement the Observer updating interface to keep its state
    consistent with the subject's.
    Store state that should stay consistent with the subject's.
    """

    def update(self, arg):
        
        self._observer_state = arg
        # ...


import os
import datetime
import logging
import threading
from queue import Queue
import time
from threading import Thread

from confluent_kafka.schema_registry.avro import AvroSerializer # Used to create the Avro serializer in conf
from confluent_kafka.schema_registry.json_schema import JSONSerializer # Used to create the JSON serializer in conf
from confluent_kafka.serialization import StringSerializer # Used to create a generic serializer in conf
from confluent_kafka import KafkaException
from confluent_kafka import SerializingProducer

from kp_fraydit.custom_types import flatten_list
from kp_fraydit.custom_errors import CustomError, OfflineError, BrokerOfflineError, RegistryOfflineError
from kp_fraydit.connections.connection import KafkaConnection, get_ip_and_port_from_string
from kp_fraydit.schema_client import SchemaEngine

# END IMPORTS ///////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# FILE LOGGER ///////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


logger = logging.getLogger(__name__)

# define file handler and set formatter
current_directory = os.path.dirname(os.path.abspath(__file__))
file_handler = logging.FileHandler(f'{current_directory}/logs/producer.log')
formatter    = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
file_handler.setFormatter(formatter)

# add file handler to logger
logger.addHandler(file_handler)


# END FILE LOGGER ///////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


# PRODUCER CLASS ////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

kConn = KafkaConnection()

class Producer:
    __counter = 0

    def __init__(self, topic_name: str, value_schema_name: str = None, key_schema_name: str = None, 
                    optional_value_args: list = [], optional_key_args: list = [], cache_schemas: bool = True
                ):

        # Check there is a connection
        if kConn.kafka_broker_listener is None: return None

        # INSTANCE COUNT AND ID /////////////////////////////////////////////////////////////////////////////////////////

        Producer.__counter += 1
        self.id = Producer.__counter + 100000

        # ESTABLISH CONNECTION //////////////////////////////////////////////////////////////////////////////////////////

        self.__broker_online = False
        self.__registry_online = False
        
        kConn.attach(self) # Attach to connection for dispatch updates
        
        # Try to connect
        tries = 0
        while not kConn.registry_online and not kConn.broker_online:
            print (f'Producer could not initialize. Attempting ({tries}) to connect...')
            time.sleep(1)
            tries += 1
            if tries == 10: 
                raise OfflineError


        # SCHEMA ATTRIBUTES AND METHODS /////////////////////////////////////////////////////////////////////////////////
        # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


        self.cache_schemas = cache_schemas # Controls round trips to schema registry
        self.__topic_name = topic_name
        

        # Set value schema default if not set
        if value_schema_name is not None:
            self.__value_schema_name = value_schema_name # Name of the value schema in kafka registry
        else:
            self.__value_schema_name = f'{self.__topic_name}-value' # if no name is specified, default.

        self.__value_fields = []
        self.__optional_value_fields = set()
        self.__required_value_fields = set()
        self.__value_record = {}
        self.__inclusive_optional_value_field_names = optional_value_args # List of optional arguments beyond the required arguments to write


        # Set key schema default is not set
        if key_schema_name is not None:
            self.__key_schema_name = key_schema_name # Name of the key schema in kafka registry
        else:
            self.__key_schema_name = f'{self.__topic_name}-key'
            
        self.__key_fields = []
        self.__optional_key_fields = []
        self.__required_key_fields = []
        self.__key_record = {}
        self.__inclusive_optional_key_field_names = optional_key_args 
        
        
        # CREATE SCHEMAS FOR SERIALIZERS ////////////////////////////////////////////////////////////////////////////////
                
        if not kConn.registry_online: # Check to see if registry is online. Raise error if not. If online, set the registry
            raise RegistryOfflineError(f'Registry self.connectionection to address: {kConn.kafka_registry_listener} failed')
        else:
            self.__schema_engine = SchemaEngine()
            
            try:
                self.__current_value_schema = self.__schema_engine.get_latest_schema(self.__value_schema_name)
                if self.__current_value_schema.schema_type == 'AVRO':
                    _value_serializer = AvroSerializer(self.__schema_engine, self.__current_value_schema.schema_str)
                elif self.__current_value_schema.schema_type == 'JSON':
                    _value_serializer = JSONSerializer(self.__current_value_schema.schema_str, self.__schema_engine)
                
            except:
                self.__current_value_schema = None
                _value_serializer = StringSerializer()

            try:
                self.__current_key_schema = self.__schema_engine.get_latest_schema(self.__key_schema_name)
                if self.__current_key_schema.schema_type == 'AVRO':
                    _key_serializer = AvroSerializer(self.__schema_engine, self.__current_key_schema.schema_str)
                elif self.__current_key_schema.schema_type == 'JSON':
                    _key_serializer = JSONSerializer(self.__current_key_schema.schema_str, self.__schema_engine)

            except:
                self.__current_key_schema = None
                _key_serializer = StringSerializer()
        

        # CREATE THE KAFKA CONFIGURATION DICTIONARY /////////////////////////////////////////////////////////////////////
        

        if not self.broker_online: # Check to see if the broker is online. Raise error if not. If online, set the broker
            raise BrokerOfflineError(kConn.kafka_broker_listener)
        else:
            ip, port = get_ip_and_port_from_string(kConn.kafka_broker_listener)
            formatted_address = f'{ip}:{port}'
            self.__kafka_broker_listener = formatted_address
            if self.__current_value_schema is None:
                producer_conf = {
                'bootstrap.servers': formatted_address,
                'value.serializer': None,
                'key.serializer': _key_serializer
                }
            elif self.__current_key_schema is None:
                producer_conf = {
                'bootstrap.servers': formatted_address,
                'value.serializer': _value_serializer,
                'key.serializer': None
                }
            elif self.__current_key_schema is None and self.__current_value_schema is None:
                producer_conf = {
                'bootstrap.servers': formatted_address,
                'value.serializer': None,
                'key.serializer': None
                }
            else:
                producer_conf = {
                'bootstrap.servers': formatted_address,
                'value.serializer': _value_serializer,
                'key.serializer': _key_serializer
                }
            
            self.__conf = producer_conf

        
        # PRODUCING ATTRIBUTES //////////////////////////////////////////////////////////////////////////////////////////
        
        
        self.__preserve_order = False
        self.__producer_thread_running = False
        self.__queue = Queue()

        self.__conf['transactional.id'] = self.id

    # END INITIALIZATION ////////////////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    # PRODUCER CONFIGURATION ////////////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    @property
    def conf(self):
        return self.__conf


    @property
    def transactional_id(self):
        return self.__transactional_id


    @transactional_id.setter
    def transactional_id(self, value):
        self.__transactional_id = value
        self.__conf['transactional.id'] = value


    @property
    def preserve_order(self):
        return self.__preserve_order


    @preserve_order.setter
    def preserve_order(self, value):
        self.__preserve_order = value


    @property
    def topic_name(self):
        return self.__topic_name


    # END PRODUCER CONFIGURATION ////////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    

    # SUBSCRIPTION METHOD TO RETRIEVE UPDATES FROM CONNECTION ///////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    def update(self, arg):
        
        self.__broker_online = arg['broker_online']
        self.__registry_online = arg['registry_online']


    # END SUBSCRIPTION METHOD TO RETRIEVE UPDATES FROM CONNECTION ///////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    

    # KAFKA CONNECTION METHODS AND ATTRIBUTES ///////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    

    @property
    def online(self):
        if self.__broker_online and self.__registry_online: return True
        
        return False


    @property
    def broker_online(self):
        return self.__broker_online


    @property
    def registry_online(self):
        return self.__registry_online


    # END KAFKA CONNECTION METHODS AND ATTRIBUTES ///////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    

    # VALUE SCHEMA ATTRIBUTES AND METHODS ///////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @property
    def value_record(self):
        return self.__value_record

    @property
    def value_schema_name(self):
        return self.__value_schema_name

    @property
    def current_value_schema(self):
        
        if self.__current_value_schema is None: return None

        if self.cache_schemas or not self.registry_online:
            return self.__current_value_schema
        else:
            self.__current_value_schema = self.__schema_engine.get_latest_schema(self.value_schema_name)
            return self.__current_value_schema
    

    def get_value_field(self, field_name:str):
        
            fields = [d for d in self.value_fields if d['name'] == field_name]
            if isinstance(fields, list): 
                if len(fields) > 0: return fields[0]
                else: return None
            else: 
                try: return fields
                except Exception as e:
                    return None
    
    @property
    def value_fields(self):

        my_fields = self.__schema_engine.get_value_fields(schema_name=self.value_schema_name)
        self.__value_fields = my_fields['value_fields']
        self.__required_value_fields = my_fields['required_value_fields']
        self.__optional_value_fields = my_fields['optional_value_fields']

        return self.__value_fields

    @property
    def value_field_names(self):
        return self.__schema_engine.get_field_names(self.value_fields)

    @property
    def required_value_fields(self):
        my_fields = self.__schema_engine.get_value_fields(schema_name=self.value_schema_name)
        self.__value_fields = my_fields['value_fields']
        self.__required_value_fields = my_fields['required_value_fields']
        self.__optional_value_fields = my_fields['optional_value_fields']

        return self.__required_value_fields


    @property
    def required_value_field_names(self):
        return self.__schema_engine.get_field_names(self.required_value_fields)

    @property
    def optional_value_fields(self):
        
        my_fields = self.__schema_engine.get_value_fields(schema_name=self.value_schema_name)
        self.__value_fields = my_fields['value_fields']
        self.__required_value_fields = my_fields['required_value_fields']
        self.__optional_value_fields = my_fields['optional_value_fields']

        return self.__optional_value_fields


    @property
    def optional_value_field_names(self):
        return self.__schema_engine.get_field_names(self.optional_value_fields)

    @property
    def inclusive_optional_value_field_names(self):
        return self.__inclusive_optional_value_field_names


    @inclusive_optional_value_field_names.setter
    def inclusive_optional_value_field_names(self,value):
        
        if isinstance(value, list): 
            scrubbed_list = []
            for v in value:
                if v in self.optional_value_field_names: 
                    scrubbed_list.append(v)
        self.__inclusive_optional_value_field_names = scrubbed_list
        

    @property
    def missing_value_field_names(self):
        all_fields = []
        all_fields.extend(self.required_value_field_names)
        all_fields.extend(self.inclusive_optional_value_field_names)
        for key in self.__value_record:
            if isinstance(self.__value_record[key], dict): # Check to see if its a record
                for subitem in self.__value_record[key]:
                    if subitem in all_fields: all_fields.remove(subitem)
            else:
                if key in all_fields: all_fields.remove(key)
        return all_fields


    def addValueArgs(self, **kwargs):
        for kwarg in kwargs:
            if kwarg in self.value_field_names: # Check to see if the field exists
                # Get the value of the kwarg and convert it to the datatype in the kafka schema
                new_val = self.convert_field_type(kwargs.get(kwarg),self.get_field_types(kwarg))
            # Do not add a none datatype. Let the schema default to null
                if new_val is not None:
                    # Check to see if key has parent
                    if 'parent' in self.get_value_field(kwarg).keys(): # field has a parent
                        # create the record
                        if self.get_value_field(kwarg)['parent'] not in self.__value_record.keys(): # test to see if parent is in value record
                            # temp_record = str(self.get_value_field(kwarg)['record'])
                            self.__value_record[self.get_value_field(kwarg)['parent']] = {}
                        # create the record
                        record_dict = {kwarg: new_val}
                        self.__value_record[self.get_value_field(kwarg)['parent']][kwarg] = new_val
                    else: # field does not have parent
                        self.__value_record[kwarg] = new_val
        self.__check_arguments()
   
    @property
    def values(self):
        if self.__current_value_schema is None:
            try: values = bytes(self.__value_record.get('value').encode('utf-8'))
            except: values = None
        else:
            values = dict(self.__value_record)
        return values


    # END VALUE SCHEMA ATTRIBUTES AND METHODS ///////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    # KEY SCHEMA ATTRIBUTES AND METHODS /////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    @property
    def key_record(self):
        return self.__key_record


    @property
    def key_schema_name(self):
        return self.__key_schema_name


    @property
    def current_key_schema(self):
        if self.__current_key_schema is None: return None
        if self.cache_schemas or not self.registry_online:
            return self.__current_key_schema
        else:
            self.__current_key_schema = self.__schema_engine.get_latest_schema(self.key_schema_name)
            return self.__current_key_schema

        
    @property
    def key_fields(self):
        
        my_fields = self.__schema_engine.get_key_fields(schema_name=self.key_schema_name)
        self.__key_fields = my_fields['key_fields']
        self.__required_key_fields = my_fields['required_key_fields']
        self.__optional_key_fields = my_fields['optional_key_fields']
        return self.__key_fields


    @property
    def key_field_names(self):
        return self.__schema_engine.get_field_names(self.key_fields)

    @property
    def required_key_fields(self):
                
        my_fields = self.__schema_engine.get_key_fields(schema_name=self.key_schema_name)
        self.__key_fields = my_fields['key_fields']
        self.__required_key_fields = my_fields['required_key_fields']
        self.__optional_key_fields = my_fields['optional_key_fields']
        return self.__required_key_fields


    @property
    def required_key_field_names(self):
        return self.__schema_engine.get_field_names(self.required_key_fields)

    @property
    def optional_key_fields(self):
        my_fields = self.__schema_engine.get_key_fields(schema_name=self.key_schema_name)      
        self.__key_fields = my_fields['key_fields']
        self.__required_key_fields = my_fields['required_key_fields']
        self.__optional_key_fields = my_fields['optional_key_fields']


        return self.__optional_key_fields


    @property
    def optional_key_field_names(self):
        return self.__schema_engine.get_field_names(self.optional_key_fields)

    @property
    def inclusive_optional_key_field_names(self):
        return self.__inclusive_optional_key_field_names


    @inclusive_optional_key_field_names.setter
    def inclusive_optional_key_field_names(self,value):
        if isinstance(value, list): 
            scrubbed_list = []
            for v in value:
                if v in self.optional_key_field_names: 
                    scrubbed_list.append(v)
        self.__inclusive_optional_key_field_names = scrubbed_list

   
    @property
    def missing_key_field_names(self):
        all_fields = []
        all_fields.extend(self.required_key_field_names)
        all_fields.extend(self.inclusive_optional_key_field_names)

        for key in self.__key_record.keys():
            if key in all_fields: all_fields.remove(key)
        return all_fields


    def addKeyArgs(self, **kwargs):
        key_fields = self.key_field_names
        for kwarg in kwargs:
            if kwarg in key_fields: # Check to see if the field exists
                # Get the value of the kwarg and convert it to the datatype in the kafka schema
                new_val = self.convert_field_type(kwargs.get(kwarg),self.get_field_types(kwarg, False))

                # Do not add a none datatype. Let the schema default to null
                if new_val is not None:
                    self.__key_record[kwarg] = new_val
                    
        self.__check_arguments()
        

    @property
    def keys(self):

        if self.__current_key_schema is None:
            try: keys = (self.__key_record.get('key'))
            except: keys = None
        else:
            keys = dict(self.__key_record)
        return keys
    
    
    # END KEY SCHEMA ATTRIBUTES AND METHODS /////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


    # FIELD PROCESSING METHODS //////////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


    def get_field_types(self, field_name, value_schema=True):
        if value_schema == True:
            results = [sub['type'] for sub in self.value_fields if sub['name'] == field_name]
        else:
            results = [sub['type'] for sub in self.key_fields if sub['name'] == field_name]
        
        return (flatten_list(results))


    def convert_field_type(self, value_to_convert, field_types_list=[]):
        
        for field_type in field_types_list:
            
            
            if field_type == 'double' or field_type == 'float' or field_type == 'number':
                try:
                    # float(value_to_convert)
                    return float(value_to_convert)
                except:
                    print (f'Error converting {value_to_convert} to float')
            elif field_type == 'long' or field_type == 'int' or field_type == 'integer':
                try:
                    return int(value_to_convert)
                except:
                    print (f'Error converting {value_to_convert} to integer')
            elif field_type == 'bytes':
                try:
                    return bytes(value_to_convert)
                except:
                    print (f'Error converting {value_to_convert} to bytes')
            elif field_type == 'boolean':
                if value_to_convert == 1 or value_to_convert == 'true' or value_to_convert == 'True' or value_to_convert == True:
                    return True
                elif value_to_convert == 0 or value_to_convert == 'false' or value_to_convert == 'False' or value_to_convert == False:
                    return False
                else:
                    print (f'Error converting {value_to_convert} to boolean')
            elif field_type =='string':
                try:
                    return str(value_to_convert)
                except:
                    print (f'Error converting {value_to_convert} to string')
            
            elif field_type == 'null':
                pass
        
        # Exit loop
        print (f'Conversion error. The value: {value_to_convert} was not of type: {field_types_list}')
        return None


    # END FIELD PROCESSING METHODS //////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


    # PRODUCE METHODS AND ATTRIBUTES ////////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////

    def __check_arguments(self):
        
        if len(self.missing_value_field_names) == 0 and len(self.missing_key_field_names) == 0: 
            if self.preserve_order: self.produce()
            else: self.queue_for_connection()

    
    def queue_for_connection(self):
        kConn.produce_queue.put([self.__topic_name, self.value_schema_name, self.key_schema_name, self.values, self.keys, self.__conf])
        
        self.__value_record.clear()  # Delete the records to start the add arguments from scratch    
        self.__key_record.clear()


    def produce(self):
        
        if not self.__producer_thread_running: 
            worker = Thread(target=self.__queue_production_loop, args=(), daemon=True)
            worker.start()        
            self.__producer_thread_running = True

        self.__queue.put([self.__topic_name, self.value_schema_name, self.key_schema_name, self.values, self.keys, self.__conf])
        self.__value_record.clear()
        self.__key_record.clear()    

    
    def __queue_production_loop(self):
        prod = SerializingProducer(self.conf)
        prod.init_transactions()
            
        while True:
            task = self.__queue.get()
            topic, value_schema, key_schema, values, keys,  conf_ = task # unpack into variables
        
            try:
                prod.begin_transaction()
                prod.produce(topic=topic, key=keys, value=values)
            
                prod.commit_transaction()
                prod.flush()
                break
            
            except KafkaException as e:
                if e.args[0].retrievable():
                    continue # retry the transaction until an abortable failure happens
                    '''
                    Upon transaction failure or unidentified failure, the transaction is requeued to be processed
                    '''
                elif e.args[0].txn_requires_abort():
                    self.abort_transaction()
                    self.__queue.put([self.__topic_name, self.value_schema_name, self.key_schema_name, values, keys, self.__conf])
                else:
                    self.abort_transaction()
                    self.__queue.put([self.__topic_name, self.value_schema_name, self.key_schema_name, values, keys, self.__conf])

    
    # END PRODUCE METHODS AND ATTRIBUTES ////////////////////////////////////////////////////////////////////////////
    # ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


# END PRODUCER CLASS ////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
# ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////



import typing

def flatten_list(list_):
    list_of_lists = []
    flattened_list = []
    list_of_lists.append(list_)
    
    while len(list_of_lists) > 0:
        for my_list in list_of_lists:
            for item in my_list:
                if isinstance(item, list):
                    list_of_lists.append(item)
                else:
                    flattened_list.append(item)
            
            list_of_lists.remove(my_list)
    return flattened_list


def key_exists(dict_to_check,key_to_check):
    if key_to_check in dict_to_check: return True
    return False

def check_int(s):
    s = str(s)
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()    


import json

from kp_fraydit.admin import AdminEngine
from kp_fraydit.connections.connection import KafkaConnection
from kp_fraydit.producers.producer import Producer
from kp_fraydit.schema_client import SchemaEngine


eng = SchemaEngine()
admin = AdminEngine()
kConn = KafkaConnection()

default_topic_creation_conf = {
    'num_partitions': '5',
    'retention_time': '-1',
    'retention_size': '-1',
}

class Connector:
    def __init__(self, source_name, topic_creation_conf=default_topic_creation_conf):
        self.__source_name = source_name

        # Check to see if the topic exists
        if not admin.topic_exists(source_name): admin.create_topic(topic_name=source_name, num_partitions=topic_creation_conf['num_partitions'], retention_time=topic_creation_conf['retention_time'], retention_size=topic_creation_conf['retention_size'])
        self.__producer = Producer(source_name)
        # self.consumer = consumer
        

    @property
    def source_name(self):
        return self.__source_name


    @property
    def producer(self):
        return self.__producer

    




import os
import json


from confluent_kafka.admin import NewTopic
from confluent_kafka.admin import AdminClient
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry import Schema

from kp_fraydit.connections.connection import KafkaConnection
from kp_fraydit.metaclasses import SingletonMeta
from .custom_errors import CustomError, RegistryOfflineError


kConn = KafkaConnection()
schema_compatibility_modes = ['FULL_TRANSITIVE', 'FULL', 'FORWARD_TRANSITIVE', 'FORWARD', 'BACKWARD', 'BACKWARD_TRANSITIVE', 'NONE']


class SchemaEngine(SchemaRegistryClient, metaclass=SingletonMeta):

    def __init__(self):
        #with ConnectionSettings() as conn:
        try:
            if not kConn.registry_online: raise RegistryOfflineError
            schema_client_conf = {'url':f'{kConn.kafka_registry_listener}'}
            self._schema_client = SchemaRegistryClient(schema_client_conf)
            super().__init__(schema_client_conf)
        
        except RegistryOfflineError as e:
            pass
        
        except Exception as e:
            raise CustomError(e)
    
    @property
    def schema_subjects(self):
        return self.get_subjects()

    
    def get_latest_schema(self, schema_subject):
        try:
            return super().get_latest_version(schema_subject).schema
        except:
            return None

    
    # GET VALUE FIELDS //////////////////////////////////////////////////////////////////////////////////////////////////
    
    '''
    Helper function that processes the schema type and sets the required, optional, and all fields for the value schema.
    This function flattens a nested Avro schema and that allows for nesting capabilities. This only works for Avro schema.
    The nested schema needs to be like this:
    fields: [
        {
            "name": "nestedField",
            "type": [
                {
                    "doc": "struture that contains nested info",
                    "name": "TestpathElement",
                    "namespace": "com.example.kafka.core",
                    "type": "record"
                    "fields": [
                        {
                            "name": "sourceNested",
                            "type": "string"
                        },
                        {
                            "name": "timestampNested",
                            "type": "long"
                        }
                    ],
                }
            ]
        }
    ]
    '''
    def get_value_fields(self, **kwargs):
        if 'schema_name' in kwargs: current_value_schema = self.get_latest_schema(kwargs.get('schema_name'))
        else: current_value_schema = kwargs.get('schema')
        value_fields = []
        all_fields = []
        required_value_fields = []
        optional_value_fields = []

    
        if current_value_schema is None:     # No schema
            value_dict = dict()
            value_dict['name'] = 'value'
            value_dict['type'] = ['string', 'float']
            value_fields = [value_dict]
            all_fields.append(value_dict)
            required_value_fields.append(value_dict)
            return {
                'value_fields': all_fields,
                'required_value_fields': required_value_fields,
                'optional_value_fields': optional_value_fields
            }

        if current_value_schema.schema_type == 'AVRO': # Avro schema
            value_fields = eval('dict('+current_value_schema.schema_str+')')['fields']
            
        elif current_value_schema.schema_type == 'JSON': # JSON schema

            json_list = []
            value_fields = eval('dict('+current_value_schema.schema_str+')')['properties']
            for value_field_name, value_field_contents in value_fields.items():
                my_dict = {}
                my_dict['name'] = value_field_name
                for k, v in value_field_contents.items():
                    if k == 'description': my_dict['doc'] = v
                    else: my_dict[k] = v
                    
                json_list.append(my_dict)
            value_fields = json_list
        
        for value_field in value_fields:
            
            required = True
            nested = False
            if isinstance(value_field['type'], list): # the type is a list
                if 'null' in value_field['type']: required = False # check for null
                # print (value_field['type'])    
                for type_ in value_field['type']: # iterate the type
                    nested = False
                    # Check for nested elements
                    if isinstance(type_, dict): # Dictionary of elements. 
                        nested = True # toggle nested for the purposes of adding the fields
                        if 'type' in type_.keys(): # type exists in dictionary
                            
                            if type_['type'] == 'record': # type is array. Get the fields from the items key
                                for element in type_['fields']: # iterate the sub elements
                                    subtype_required = True # toggle on required for subtype
                                    element['record'] = type_['name'] # add the parent
                                    element['parent'] = value_field['name']
                                    # Add the field to all fields
                                    all_fields.append(element)
                                    # Add the field to either required or optional
                                    if isinstance(element['type'], list):
                                        
                                        for subtype_ in element['type']: # iterate type that is list
                                            if subtype_ == 'null': subtype_required = False
                                    else:
                                        if element['type'] == 'null': subtype_required = False
                                    # Set if optional or required
                                    if subtype_required and required: required_value_fields.append(element)
                                    else: optional_value_fields.append(element)
                        else: # the type is a list. there is a dictionary in the type. type does not exist in in the keys
                            pass
                    else: # the type is a list. there is a dictionary in the type
                        pass

            else: # the type is not a list. It is not nested
                if value_field == 'null': required = False
            
            if not nested:
                all_fields.append(value_field)
                if required: required_value_fields.append(value_field)
                else: optional_value_fields.append(value_field)
    
        return {
            'value_fields': all_fields,
            'required_value_fields': required_value_fields,
            'optional_value_fields': optional_value_fields
        }

    # END GET VALUE FIELDS //////////////////////////////////////////////////////////////////////////////////////////////


    # GET KEY FIELDS ////////////////////////////////////////////////////////////////////////////////////////////////////

    def get_key_fields(self, **kwargs):
        if 'schema_name' in kwargs: current_key_schema = self.get_latest_schema(kwargs.get('schema_name'))
        else: current_key_schema = kwargs.get('schema')

        
        # current_key_schema = schema
        key_fields = []
        all_fields = []
        required_key_fields = []
        optional_key_fields = []

        # Return value for no key
        if current_key_schema is None:
            key_dict = dict()
            key_dict['name'] = 'key'
            key_dict['type'] = ['null', 'string']
            key_fields = [key_dict]
            all_fields.append(key_dict)
            optional_key_fields.append(key_dict)
            return {
                'key_fields': all_fields,
                'required_key_fields': required_key_fields,
                'optional_key_fields': optional_key_fields
            }

        # process avro schema
        elif current_key_schema.schema_type == 'AVRO': 
            key_fields = eval('dict('+current_key_schema.schema_str+')')['fields']

        # process json schema
        elif current_key_schema.schema_type == 'JSON': 

            json_list = []
            key_fields = eval('dict('+current_key_schema.schema_str+')')['properties']
            for key_field_name, key_field_contents in key_fields.items():
                my_dict = {}
                my_dict['name'] = key_field_name
                for k, v in key_field_contents.items():
                    if k == 'description': my_dict['doc'] = v
                    else: my_dict[k] = v
                    
                json_list.append(my_dict)
            key_fields = json_list

        """
        The key fields have been processed according to the schema. 
        """
        all_fields = []

        for key_field in key_fields:

            all_fields.append(key_field)
            
            not_required = False
            if type(key_field['type']) is list:
                for type_ in key_field['type']:
                    if type_ == 'null': not_required = True
            else: 
                if key_field == 'null': not_required = True

            if not_required: optional_key_fields.append(key_field)
            else: required_key_fields.append(key_field)

        return {
            'key_fields': all_fields,
            'required_key_fields': required_key_fields,
            'optional_key_fields': optional_key_fields
        }

    # END GET KEY FIELDS //////////////////////////////////////////////////////////////////////////////////////////////

    def get_field_names(self, list_):
        
        if not isinstance(list_, list): return []
        my_list = list()
        for field in list_:
            
            my_list.append(field['name'])
        return my_list

    
    def schema_exists(self, schema_name):
        try:
            value = self.get_latest_schema(schema_name)

            if value is None: return False
        except:
            return False

        return True


    def validate_schema(self, subject_name, schema_name, base_schema_source):

        # VALIDATE VALUE SCHEMA ////////////////////////////////////////////////

        # Load the json data structure
        with open(base_schema_source) as json_file:
            data = json.load(json_file)
            base_value_fields = data['fields']
            base_value_field_names = self.get_field_names(base_value_fields)


        # set the head of the schema
        new_schema_dict = {}
        new_schema_dict['type'] = data['type']
        schema_namespace = f'{data["namespace"]}.{subject_name}'
        new_schema_dict['namespace'] = schema_namespace
        new_schema_dict['doc'] = data['doc']
        new_schema_dict['name'] = subject_name

        if self.schema_exists(schema_name):
            current_schema = self.get_latest_schema(schema_name)
            current_schema_dict = json.loads(current_schema.schema_str)
            

            values = self.get_value_fields(schema=current_schema)
            new_values = values['value_fields']
            
            # Validate that the existing fields are good
            for field in new_values:
                if field['name'] in base_value_field_names:
                    # Get the base field
                    base_field = [f for f in base_value_fields if f['name'] == field['name']][0]
                    
                    # Check the field attributes and make sure they are identical. If not, delete the field
                    for key, value in field.items():
                        if not base_field.get(key) == value:
                            new_values.remove(field)
                            break

            # Add fields that are missing
            print (set(base_value_field_names))
            print (set(self.get_field_names(new_values)))
            missing_value_fields = set(base_value_field_names) - set(self.get_field_names(new_values))
            print (f'Missing fields: {missing_value_fields}')
            
            # Count missing fields. If none, exit with no change
            if not len(missing_value_fields):
                return False

            for field in missing_value_fields:
                field_dict = [d for d in base_value_fields if d['name'] == field]
                new_values.append(field_dict[0])
            
            
            new_schema_dict['fields'] = new_values

        else: # Create a schema from base
            print ('Creating new schema')
            new_schema_dict = data
            new_schema_dict['namespace'] = schema_namespace
        

        # Format namespace and create the new Schema object
        new_schema_dict['namespace'] = f'{schema_namespace}' # Set namespace from previous value
        formatted_schema_str = str(new_schema_dict).replace("'", '"') # replace single quotes with double quotes
        # formatted_schema_str = str(new_schema_dict)
        new_schema = Schema(schema_str=str(formatted_schema_str),schema_type='AVRO')

        # Try to update the schema. On Failure, iterate through compatiblity modes and try to change
        while True:
            try:
                # Register the schema
                self.register_schema(schema_name, new_schema)
                return True
            except:
                pass
                # print ('failed')
            
            for count, comp in enumerate(schema_compatibility_modes):
                try:
                    print (f'Compare: {comp} Count: {count}')
                    self.set_compatibility(schema_name, level=comp)
                    # Register the schema
                    self.register_schema(schema_name, new_schema)
                    return True
                except:
                    if count == len(schema_compatibility_modes): return False




import os
import time
import concurrent.futures
import socket

from ..custom_types import key_exists, check_int


def check_host_port(host: str, port: int, tag='', tries=3, sleep_interval=1):
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    a_socket.settimeout(3)
    location = (host, port)
    actual_tries = 0
    while tries > actual_tries:
        
        try:
            result_of_check = a_socket.connect_ex(location)
            if result_of_check == 0: return True, host, port, tag, time.time()
            actual_tries += 1
            time.sleep(sleep_interval)
        
        except:
            actual_tries += 1
            time.sleep(sleep_interval)
                
    return False, host, port, tag


def get_ip_and_port_from_string(ip_string: str):
    
    stripped_string = ip_string.strip()
    if stripped_string[:7] == 'http://': stripped_string = ip_string[7:]
    if stripped_string[:8] == 'https://': stripped_string = ip_string[8:]
    
    if len(stripped_string.split(':')) > 1: return stripped_string.split(':')[0], stripped_string.split(':')[1]
    else: return stripped_string, ''
    



def valid_address_format(url):
    stripped_string = url.strip()
    port = stripped_string.split(':')[-1:][0]
    if not check_int(port): return False
    
    if stripped_string[:7] == 'http://' or stripped_string[:8] == 'https://': 
        return True
    
    return False

class ConnectionSettingsMeta(type):
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)

        return cls._instance

class ConnectionSettings(metaclass=ConnectionSettingsMeta):
    
    def __init__(self, reping_time=3):
    
        self._kafka_broker_listener = os.environ.get('KAFKA_BROKER_LISTENER')
        self._kafka_registry_listener = os.environ.get('KAFKA_REGISTRY_LISTENER')
        self._status = {}
        self._reping_time = reping_time # Sets the time that a ping is considered valid
        # cls.check_settings()
        

    @property
    def kafka_broker_listener(self):
        return self._kafka_broker_listener


    @kafka_broker_listener.setter
    def kafka_broker_listener(self, value):
        ip, port = get_ip_and_port_from_string(value)
        formatted_address = f'{ip}:{port}'
        os.environ['KAFKA_BROKER_LISTENER'] = formatted_address
        self._kafka_broker_listener = formatted_address
        
        
    @property
    def kafka_registry_listener(self):
        return self._kafka_registry_listener


    @kafka_registry_listener.setter
    def kafka_registry_listener(self, value):
        if not valid_address_format(value):
            print (f'Address supplied ({value}) for kafka registry listener is not valid. Please format with http or https preceding address and a port number')
            return
        os.environ['KAFKA_REGISTRY_LISTENER'] = value
        self._kafka_registry_listener = value
        
        
KafkaConnectionSettings = ConnectionSettings()



import logging
import os
import datetime

from kp_fraydit.datetime_functions import today_as_string


# setup the logger
logger = logging.getLogger(__name__)

# define file handler and set formatter
current_directory = os.path.dirname(os.path.abspath(__file__))
file_handler = logging.FileHandler(f'{current_directory}/logs/errors_{today_as_string()}.log')
formatter    = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
file_handler.setFormatter(formatter)

# add file handler to logger
logger.addHandler(file_handler)

# Set logger level to error
logger.setLevel(logging.ERROR)

class CustomError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
            if len(args) > 1: self.error_type = args[1]
        else:
            self.message = None

        print(self)
    
    def __str__(self):
    
        if hasattr(self, 'message') and hasattr(self, 'error_type'):
            return f'CustomError: {self.error_type}: {self.message}'
        elif hasattr(self, 'message'):
            return f'CustomError: {self.message}'
        else:
            return 'CustomError has been raised'


class OfflineError(CustomError):
    def __init__(self):
        self.message = f'Kafka if unreachable'
        self.error_type = 'OfflineError'
        super().__init__(self.message, self.error_type)
        logger.error(self.message)


class BrokerOfflineError(CustomError):
    def __init__(self, address):
        self.address = address
        self.message = f'Kafka broker unreachable at address: {self.address}'
        self.error_type = 'BrokerOfflineError'
        super().__init__(self.message, self.error_type)
        logger.error(self.message)


class RegistryOfflineError(CustomError):
    def __init__(self, address):
        self.address = address
        self.message = f'Kafka registry unreachable at address: {self.address}'
        self.error_type = 'RegistryOfflineError'
        super().__init__(self.message, self.error_type)
        logger.error(self.message)